(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/lexer.lex,v
   Comment: Lexer for vsr/vsp file. Originally by Jorn Lind Nielsen.
            Massively modified by Andrzej Wasowski. New languages added (RL).
   Release: scope-0-15_final

     Copyright (C) -2001 Jorn Lind Nielsen
     Copyright (C) 2001-2004 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   * 2004/01

       Lexer support added based on original IAR lexer files. IAR had made 
       them available to me as a form of specification. I have not 
       incorporated the original files, but instead incorporated changes into 
       my lexer/parser files (as we have a different structure. This was
       needed to preserve my completely different structure of the files
       and... copyrights.
      

   lexer.lex,v 1.12 2004/05/05 21:29:05 wasowski Exp
*)
{
  open Lexing parser;
  open StringCvt;

  (* error handling primitives *)

  fun mkLoc lexbuf = Location.Loc(getLexemeStart lexbuf, getLexemeEnd lexbuf)
 
  fun lexerError lexbuf msg = Err.lexicalErr (mkLoc lexbuf) msg


  (* value constructors *)

  fun mkchr s = T_charval (String.extract (s,1,SOME 2))

  val stringStart   = ref 0;
  val stringContent = ref ""

  fun stringNotClosed lexbuf =
     Err.lexicalErr (Location.Loc (!stringStart, getLexemeEnd lexbuf))
   		      "Nonterminated string.";

  fun getDecConstant lexbuf =
     case Int.fromString(getLexeme lexbuf) of
         SOME c => T_decval c
       | NONE   => lexerError lexbuf "Constant out of range"

  fun getHexConstant lexbuf =
     case Int.scan HEX Substring.getc (Substring.all (getLexeme lexbuf)) of
         SOME (i,rest) => if Substring.isEmpty(rest) 
         			then T_hexval i
 				else lexerError lexbuf "Internal error."
       | NONE   => lexerError lexbuf "Constant out of range"

  fun a_keyword s lexbuf =
    case s of
        "@visualSTATE"       => T_visualSTATE
      | "@RuleBase"          => T_RuleBase
      | "@Project"           => T_Project
      | "@Directives"        => T_Directives
      | "@ExternalInterface" => T_ExternalInterface
      | "@ExternalEvent"     => T_ExternalEvent
      | "@ExternalAction"    => T_ExternalAction
      | "@ExternalVariable"  => T_ExternalVariable
      | "@Constant"          => T_Constant
      | "@InternalInterface" => T_InternalInterface
      | "@InternalVariable"  => T_InternalVariable
      | "@EventGroup"        => T_EventGroup
      | "@Signal"	      => T_Signal
      | "@Implementation"    => T_Implementation
      | "@StateMachine"      => T_StateMachine
      | "@TopStateMachine"   => T_TopStateMachine
      | "@Event"	      => T_Event
      | "@Action"	      => T_Action
      | "@Timer"             => T_Timer
      | "@State"             => T_State
      | "@Initial"	      => T_Initial
      | "@StateRegion"       => T_StateRegion
      | "@History"           => T_History
      | "@DeepHistory"	      => T_DeepHistory
      | "@Entry"             => T_Entry
      | "@Exit"              => T_Exit
      | "@Do"                => T_Do
      | "@InternalRule"      => T_InternalRule
      | "@TransitionRule"    => T_TransitionRule
      | "@Then"	      => T_Then
      | "@End"		      => T_End
      | "@Final"	      => T_Final
      | _ => lexerError lexbuf "Unknown @keyword";

  fun keyword s lexbuf =
    case s of
        "VS_VOID"    => T_typeid vsr.VoidT 
      | "VS_VOIDPTR" => T_typeid vsr.VoidPtrT
      | "VS_BOOL"    => T_typeid vsr.BoolT 
      | "VS_CHAR"    => T_typeid vsr.CharT
      | "VS_UCHAR"   => T_typeid vsr.UCharT
      | "VS_INT"     => T_typeid vsr.IntT 
      | "VS_INT8"    => T_typeid vsr.Int8T 
      | "VS_INT16"   => T_typeid vsr.Int16T 
      | "VS_INT32"   => T_typeid vsr.Int32T 
      | "VS_INT64"   => T_typeid vsr.Int32T before Err.warning 
       			"VS_INT64 not supported: changing for VS_INT32"
      | "VS_UINT"    => T_typeid vsr.UIntT 
      | "VS_UINT8"   => T_typeid vsr.UInt8T 
      | "VS_UINT16"  => T_typeid vsr.UInt16T 
      | "VS_UINT32"  => T_typeid vsr.Int32T before Err.warning
      			"VS_UINT32 not supported: changing for VS_INT32"
      | "VS_UINT64"  => T_typeid vsr.Int32T before Err.warning
      			"VS_UINT64 not supported: changing for VS_INT32"
      | "VS_FLOAT"   => T_typeid vsr.FloatT 
      | "VS_DOUBLE"  => T_typeid vsr.DoubleT
      | _            => T_id s   
}                                 

rule Token = parse

  "@visualSTATE "[`0`-`9``-`]*`;` { T_visualSTATEdir( getLexeme lexbuf ) }
| "@Date "[`0`-`9``-`]*`;`        { T_Datedir       ( getLexeme lexbuf ) }
| "@Time "[`0`-`9``:`]*`;`        { T_Timedir       ( getLexeme lexbuf ) }
| "@Field"			  { T_Fielddir }

| `@`[`a`-`z``A`-`Z`][`a`-`z``A`-`Z``0`-`9``_`]*
				  { a_keyword (getLexeme lexbuf) lexbuf }


| "&&" { T_and      }
| "||" { T_or       }
| "&"  { T_band     }
| "|"  { T_bor      }
| "~"  { T_bnot     }
| "<"  { T_lt       }
| "<=" { T_lte      }
| ">"  { T_gt       }
| ">=" { T_gte      }
| "==" { T_equ      }
| "!=" { T_neq      }
| "^"  { T_hat      }
| "["  { T_lsqbr    }
| "]"  { T_rsqbr    }
| "[(" { T_lbi      } 
| ")]" { T_rbi      }
| "[|" { T_lbe      }
| "|]" { T_rbe      }
| "[[" { T_tagbrl   }
| "]]" { T_tagbrr   }
| ":"  { T_colon    }
| "*"  { T_mul      }
| "%"  { T_percent  }
| "/"  { T_slash    }
| "#"  { T_hashmark }
| "-"  { T_sub      }
| "+"  { T_add      }
| "."  { T_dot      }
| "!"  { T_not      }
| "("  { T_lpar     }
| ")"  { T_rpar     }
| "{"  { T_lbrace   }
| "}"  { T_rbrace   }
| ";"  { T_semi     }
| ","  { T_comma    }
| "="  { T_assign   }
| ">>" { T_shr      }
| "<<" { T_shl      }


| (`'`)[^`"`](`'`)    	{ mkchr (getLexeme lexbuf) }
| (`'`)(`\\`)[`a` `b` `f` `n` `r` `t` `v` `\\` `'` `"` `?`](`'`)
			{ mkchr (getLexeme lexbuf) }
| ("'\\x'")[`0`-`9` `A`-`F` `a`-`f`][`0`-`9` `A`-`F` `a`-`f`]
			{ mkchr (getLexeme lexbuf) }



| [`a`-`z``A`-`Z``_`][`a`-`z``A`-`Z``0`-`9``_`]*
			{ keyword (getLexeme lexbuf) lexbuf }
| [`0`-`9`]+            { getDecConstant lexbuf }
| (`0`)(`x`|`X`)([`0`-`9``A`-`F``a`-`f`]+)
			{ getHexConstant lexbuf }
| ((`+`|`-`)?) ([`0`-`9`]+)(`.`)([`0`-`9`]+)(((`e`|`E`)((`+`|`-`)?)([`0`-`9`]+))?) 
			{ T_realval( getLexeme lexbuf ) }
| "\""			{ stringStart := getLexemeStart lexbuf;
                          stringContent := ""; 
			  ParseString lexbuf }
| [` ` `\t` `\n` `\r`]  { Token lexbuf }
| eof                   { T_eof }
| _                     { lexerError lexbuf "Illegal symbol in input" }


and ParseString = parse
      "\""		{ T_string( !stringContent ) }
    | "\\\""		{ stringContent := (!stringContent)^(getLexeme lexbuf); ParseString lexbuf }
    | (eof | `\^Z`)     { stringNotClosed lexbuf }
    | _			{ stringContent := (!stringContent)^getLexeme(lexbuf); ParseString lexbuf }

;
