(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/vsr.sml,v
   Comment: Abstract syntax for vsr files.  File originates from Jorn Lind
            Nilsen, massively modified (reimplemented) by Andrzej Wasowski.
   Release: scope-0-15_final

     Copyright (C) 2001-2002 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   vsr.sml,v 1.23 2002/10/24 20:17:10 wasowski Exp
*)

  open Types
  open Prim

  type symbol = Symbol.symbol 

  (* fully qualified state name *)
  
  type stpath =  symbol list 

  (* instantiation of state machine *)

  type instance = symbol * stpath list option 

  (* literals with formatting properties *)

  datatype literal = 
	  RealConst of string 
	| DecConst  of int 
	| HexConst  of int 
	| CharConst of string

  datatype tagfield = 
	  PlainTag     of string * string 
	| SysName      of int    * symbol
	| FileName     of int    * int    * symbol
	| FilePath     of int    * int    * symbol
	| CallTag      of int    * int    * instance 
	| Explanation  of symbol
	| ProjectExpl  of symbol
	| ProjectName  of symbol
	| SystemExpl   of int * symbol
	| FileExpl     of int * int * symbol
	| FreeTextText of int * symbol
	| SystemNote   of int * int * symbol
	| CenterHeader of symbol
	| RightHeader  of symbol
	| LeftFooter   of symbol
	| CenterFooter of symbol
	| RightFooter  of symbol
	| ActionFile   of symbol
	| CoderCmd     of int * symbol			(* system, options *)

  type tag = tagfield list 
  type range = ( literal * literal ) option

  type parSpec = simplety * symbol 

  (* restricted version of C prototypes *)
  type funSpec = simplety * symbol * parSpec list option

  (* used for events and singals: entity with name and typed fields *)
  type recSpec = symbol * parSpec list option

  (* external/internal events *)

  type eventSpec = recSpec * tag option
  type eventInstance = symbol * symbol list option 


   
  datatype actionSpec =
	  ActionSpec of funSpec * tag option
	| ActionTimerSpec of funSpec * tag option

  type constSpec = symbol * literal * tag option

  datatype varSpec = 
	  VarSimple of simplety * symbol * range * literal option * tag option
		    (* type * id     * range * initial value  * tag *)
	| VarArray  of simplety * symbol * literal    * range * literal list option * tag option
		    (* type * id     * array size * range * initial values      * tag *)

  (* towards internal interface *)
  
  type groupSpec = symbol * symbol list option * tag option

  (* interface *)

  type interface = {
  
	events    : eventSpec  list,
	extvars   : varSpec    list,
	actions   : actionSpec list,
	constants : constSpec  list,

	intvars   : varSpec   list,
	groups    : groupSpec list,
	signals   : eventSpec list
  }

  (* states *)

  datatype stateType = 
	  InitialState 
	| HistoryState 
	| DeepHistoryState 
	| NormalState 
	| FinalState
 
  (* name, optional params *)

  type doSpec = instance * tag option

  (* states *)

  type state = { stateType : stateType,
		 name      : symbol,
		 refersTo  : instance option,
		 tag       : tag option }

  type region = symbol * symbol list option * tag option

  datatype guard =
	  GuardPos of stpath
	| GuardNeg of stpath
	| GuardExpr of expr * bool (* true for external *)

  (* some "real" programming language staff - expressions *)

  and var = 
	  SimpleVar of symbol 
	| SubscriptVar of symbol * expr

  and expr =
	  ExprLiteral of literal
	| ExprFunCall of funCall
	| ExprVar     of var
	| ExprPar     of expr
	| ExprUnary   of unop * expr
	| ExprBinary  of expr * binop * expr
 
  and funCall = FunCall of symbol * expr list

  datatype action =
	  ActionNextState of stpath
	| ActionFunCall of funCall 
	| ActionSignal of symbol * expr list option
	| ActionAssign of var * expr * bool (* true for external *)


  type rule = { event     : eventInstance,
		condition : guard  list,
		actions   : action list,
		tag       : tag option }

  type stateMachineDef = {
  
	name		: symbol,
	param		: symbol list, 
	interface       : interface,
	states		: state list,
	regions		: region list,
	entryRules	: (action list * tag option) list,
	exitRules	: (action list * tag option) list,
	doRules         : doSpec list,
	intRules	: rule list,
	trsRules	: rule list
  }

  
  
  datatype stateMachine =
	  StateMachine    of stateMachineDef * tag option
	| TopStateMachine of stateMachineDef * tag option

	

  datatype directive = 
	  Dir_visualSTATE of string 
	| Dir_Date        of string
	| Dir_Time        of string
	| Dir_Field       of tag option

  (* vsr files *)

  type program = {
  
	filepath   : symbol,
	directives : directive list,	(* corresp. to <directives>       *)
	interface  : interface,
	machines   : stateMachine list,
	top        : stateMachine	(* coresp. to TopStateMachine    *)

  }

  (* vsp files *)

  type component = int * int * program option
  type project = {
	
	filepath   : symbol,
	directives : directive list,
	interface  : interface,
	files      : (int * int * symbol  ) list,
	programs   : (int * int * program ) list,
	calls      : (int * int * instance) list,
	sysnames   : (int * symbol) list
  }

  (* fake type which will allow to use the same 
     typechecking function for both *)

  datatype file = 
	  Project of project 
	| Program of program
	

